#include "gsa.h"

int fake_animation = 0;

int read_gsa(gsa_t &gsa, gsm_t &gsm, char *filename) {
  unsigned short i, j;
  FILE *file = open_read(filename);
  if (! file) return 1;

  char magic[16];
  if (read_arbitrary(file, magic, sizeof(magic))) return 2;
  if (str_diff(magic, GSA_MAGIC)) {
    fprintf(stderr, "%s is not a GSM\n", filename);
    return 1;
  }
  /* read version */
  if (read_uint16(file, &gsa.version)) return 2;
  /* skip unknown and skeleton shit (short, short, short, short, short) */
  if (skip(file, 10)) return 4;
  /* read skeleton crap */
  if (read_uint16(file, &gsa.name_length)) return 3;
  gsa.name = new char[gsa.name_length + 1];
  if (read_arbitrary(file, gsa.name, gsa.name_length)) return 4;
  gsa.name[gsa.name_length] = '\0';

  /* read number of bones */
  if (read_uint16(file, &gsa.num_bones)) return 5;
  gsa.bones = new bone_t[gsa.num_bones];

  for (i = 0; i < gsa.num_bones; i++) gsa.bones[i].animated = 0;

  int ret = read_bones(file, gsa.num_bones, gsa.bones);
  if (ret) return ret;

  if (gsa.num_bones != gsm.num_bones) {
    printf("Number of bones in animation (%u) doesn't match mesh (%u)\n", gsa.num_bones, gsm.num_bones);
    return 1;
  }

  /* skip marker */
  if (skip(file, 2)) return 6;

  /* read number of animated bones */
  if (read_uint16(file, &gsa.num_animated)) return 7;
  if (gsa.num_animated) {
    gsa.animated = new unsigned short[gsa.num_animated];
    /* read animated bones */
    for (i = 0; i < gsa.num_animated; i++) {
      if (read_uint16(file, &gsa.animated[i])) return 8;
      if (gsa.animated[i] >= gsa.num_bones) return 8;
      gsa.bones[gsa.animated[i]].animated = 1;
    }
  }
  else {
    gsa.num_animated = gsa.num_bones;
    gsa.animated = new unsigned short[gsa.num_animated];
    for (i = 0; i < gsa.num_animated; i++) {
      gsa.animated[i] = i;
      gsa.bones[i].animated = 1;
    }
  }
  
  /* read number of frames */
  if (read_uint16(file, &gsa.num_frames)) return 9;
  if (gsa.num_frames) gsa.frames = new frame_t[gsa.num_frames];

  for (i = 0; i < gsa.num_bones; i++) {
    bone_t &bone = gsa.bones[i];
    bone_t &meshbone = gsm.bones[i];
    bone.orientations = new sgVec4[gsa.num_frames];
    bone.locals = new sgMat4[gsa.num_frames];
    bone.worlds = new sgMat4[gsa.num_frames];
    /* copy default local rotation */
    for (j = 0; j < gsa.num_frames; j++) {
      sgCopyVec4(bone.orientations[j], meshbone.orientation);
    }
  }

  /* read frames */
  for (i = 0; i < gsa.num_frames; i++) {
    frame_t &frame = gsa.frames[i];
    if (read_vec3(file, &frame.position)) return 10;
    if (read_vec4(file, &frame.orientation)) return 11;
    if (read_vec3(file, &frame.root_position)) return 12;
    /* read animated bones */
    for (j = 0; j < gsa.num_animated; j++) {
      unsigned short bone_id = gsa.animated[j];
      bone_t &bone = gsa.bones[bone_id];
      if (read_vec4(file, &bone.orientations[i])) return 13;
    }

    /* calculate matrices */
    for (j = 0; j < gsa.num_bones; j++) {
      bone_t &bone = gsa.bones[j];
      bone_t &meshbone = gsm.bones[j];
      /* convert frame rotation to matrix */
      sgQuatToMatrix(bone.locals[i], bone.orientations[i]);
      /* add MESH's translation to the matrix */
      sgCopyVec3(bone.locals[i][3], meshbone.position);
      /* multiply */
      if (bone.parent == 0xffff) sgCopyMat4(bone.worlds[i], bone.locals[i]);
      else sgMultMat4(bone.worlds[i], gsa.bones[bone.parent].worlds[i], bone.locals[i]);
    }

    /* read number of events */
    if (gsa.version % 16 < 2) {
      frame.num_events = 0;
      continue;
    }
    if (read_uint16(file, &frame.num_events)) return 14;
    if (frame.num_events) frame.events = new event_t[frame.num_events];
    /* read events */
    for (j = 0; j < frame.num_events; j++) {
      event_t &event = frame.events[j];
      if (read_uint16(file, &event.name_length)) return 15;
      event.name = new char[event.name_length];
      if (read_arbitrary(file, event.name, event.name_length)) return 16;
      /* length includes trailing null */
      event.name_length--;
    }
  }

  close_file(file);

  return 0;
}

void create_fake_animation(gsm_t &gsm, gsa_t &gsa) {
  unsigned short i;

  /* Signal that we have a fake animation */
  fake_animation = 1;

  /* create a fake standing animation using default pose frame */
  gsa.num_frames = 1;
  gsa.name_length = gsm.name_length;
  gsa.name = new char[gsa.name_length + 1];
  memmove(gsa.name, gsm.name, gsa.name_length + 1);

  gsa.num_bones = gsm.num_bones;
  gsa.num_animated = gsa.num_bones;
  gsa.animated = new unsigned short[gsa.num_animated];
  for (i = 0; i < gsa.num_animated; i++) gsa.animated[i] = i;

  gsa.frames = new frame_t[gsa.num_frames];
  for (i = 0; i < gsa.num_frames; i++) {
    frame_t &frame = gsa.frames[i];
    memset(frame.position, 0, sizeof(sgVec3));
    memset(frame.root_position, 0, sizeof(sgVec3));
    sgMakeIdentQuat(frame.orientation);
    frame.num_events = 0;
  }

  gsa.bones = new bone_t[gsa.num_bones];
  for (i = 0; i < gsa.num_bones; i++) {
    bone_t &bone = gsa.bones[i];
    bone_t &meshbone = gsm.bones[i];
    bone.parent = meshbone.parent;
    bone.name_length = meshbone.name_length;
    bone.name = new char[bone.name_length + 1];
    memmove(bone.name, meshbone.name, bone.name_length);
    bone.name[bone.name_length] = '\0';
    bone.animated = 1;
    bone.orientations = 0;
    bone.locals = new sgMat4[gsa.num_frames];
    bone.worlds = new sgMat4[gsa.num_frames];
    sgCopyMat4(bone.locals[0], meshbone.local);
    sgCopyMat4(bone.worlds[0], meshbone.world);
  }
}

void cleanup_events(unsigned short num_events, event_t *events) {
  if (! num_events || ! events) return;
  for (unsigned short i = 0; i < num_events; i++) {
    event_t &event = events[i];
    if (event.name_length && event.name) delete[] event.name;
    event.name_length = 0; event.name = 0;
  }
  delete[] events;
}

void cleanup_frames(unsigned short num_frames, frame_t *frames) {
  if (! num_frames || ! frames) return;
  for (unsigned short i = 0; i < num_frames; i++) {
    frame_t &frame = frames[i];
    cleanup_events(frame.num_events, frame.events);
    frame.num_events = 0; frame.events = 0;
  }
  delete[] frames;
}

void cleanup_gsa(gsa_t *gsa) {
  if (gsa->name_length && gsa->name) delete[] gsa->name;
  gsa->name_length = 0; gsa->name = 0;
  if (gsa->num_animated && gsa->animated) delete[] gsa->animated;
  gsa->num_animated = 0; gsa->animated = 0;
  cleanup_bones(gsa->num_bones, gsa->bones, gsa->num_frames);
  gsa->num_bones = 0; gsa->bones = 0;
  cleanup_frames(gsa->num_frames, gsa->frames);
  gsa->num_frames = 0; gsa->frames = 0;
  delete gsa;
}
