#include "gsm.h"

int read_gsm(gsm_t &gsm, char *filename) {
  unsigned short i, j;
  int k;
  FILE *file = open_read(filename);
  if (! file) return 1;

  char magic[16];
  if (read_arbitrary(file, magic, sizeof(magic))) return 2;
  if (str_diff(magic, GSM_MAGIC)) {
    fprintf(stderr, "%s is not a GSM\n", filename);
    return 1;
  }
  /* read version shit */
  if (read_uint16(file, &gsm.version)) return 2;
  /* skip skeleton shit (short, short) */
  if (skip(file, 4)) return 4;
  /* read skeleton crap */
  if (read_uint16(file, &gsm.name_length)) return 3;
  gsm.name = new char[gsm.name_length + 1];
  if (read_arbitrary(file, gsm.name, gsm.name_length)) return 4;
  gsm.name[gsm.name_length] = '\0';

  /* read number of bones */
  if (read_uint16(file, &gsm.num_bones)) return 5;
  gsm.bones = new bone_t[gsm.num_bones];

  int ret = read_bones(file, gsm.num_bones, gsm.bones);
  if (ret) return ret;

  /* read number of assignments */
  if (read_uint16(file, &gsm.num_assignments)) return 12;
  gsm.assignments = new assignment_t[gsm.num_assignments];

  /* read assignments */
  gsm.need_hack = 0;
  for (i = 0; i < gsm.num_assignments; i++) {
    assignment_t &assignment = gsm.assignments[i];
    if (read_uint16(file, &assignment.bone_id)) return 13;
    if (assignment.bone_id == 0xeeee) {
#if 0
      for (j = i; j < gsm.num_assignments; j++) {
        /* no-op; delete[] will cleanup safely */
      }
#endif
      gsm.need_hack = gsm.num_assignments;
      gsm.num_assignments = i;
      break;
    }
    else if (assignment.bone_id >= gsm.num_bones) return 13;
    if (read_uint16(file, &assignment.first_position)) return 14;
    if (read_uint16(file, &assignment.num_positions)) return 15;
    if (skip(file, 2)) return 16;
  }

  /* blended position header */
  if (gsm.need_hack) {
    if (read_uint16(file, &gsm.num_blended)) return 17;
    if (skip(file, 4)) return 17;
  }
  else gsm.num_blended = 0;

  /* read number of positions */
  if (read_uint16(file, &gsm.num_positions)) return 18;
  gsm.positions = new position_t[gsm.num_positions];

  /* read the positions */
  for (i = 0; i < gsm.num_positions; i++) {
    position_t &position = gsm.positions[i];
    sgMakeIdentMat4(position.local);
    sgVec3 xyz;
    if (read_vec3(file, &xyz)) return 19;
    sgCopyVec3(position.local[3], xyz);
  }
  for (i = 0; i < gsm.num_assignments; i++) {
    assignment_t &assignment = gsm.assignments[i];
    if (assignment.first_position >= gsm.num_positions + gsm.num_blended) return 15;
    if (assignment.first_position + assignment.num_positions > gsm.num_positions + gsm.num_blended) return 15;
  }

  /* set bone IDs */
  for (i = 0; i < gsm.num_assignments; i++) {
    assignment_t &assignment = gsm.assignments[i];
    for (j = 0; j < assignment.num_positions; j++) {
      gsm.positions[j + assignment.first_position].bone_id = assignment.bone_id;
    }
  }

  /* read number of blended positions */
  if (read_uint16(file, &gsm.num_blended)) return 20;
  if (read_uint32(file, &gsm.blended_length)) return 21;
  if (gsm.num_blended) {
    gsm.blended = new blended_t[gsm.num_blended];

    /* read blended positions */
    for (i = 0; i < gsm.num_blended; i++) {
      blended_t &blended = gsm.blended[i];
      if (read_uint16(file, &blended.num_positions)) return 22;
      blended.positions = new unsigned short[blended.num_positions];
      blended.weights = new float[blended.num_positions];
      sgVec3 result = { 0.0f, 0.0f, 0.0f };
      for (j = 0; j < blended.num_positions; j++) {
        if (read_uint16(file, &blended.positions[j])) return 23;
        if (read_float(file, &blended.weights[j])) return 24;
        if (blended.positions[j] >= gsm.num_positions) return 24;
        for (k = 0; k < 3; k++) {
          result[k] += gsm.positions[blended.positions[j]].world[3][k] * blended.weights[j];
        }
      }
    }
  }

  /* read number of uvs */
  if (read_uint16(file, &gsm.num_uvs)) return 25;
  gsm.uvs = new uv_t[gsm.num_uvs];

  /* read uvs */
  for (i = 0; i < gsm.num_uvs; i++) {
    uv_t &uv = gsm.uvs[i];
    if (read_arbitrary(file, &uv, sizeof(uv_t))) return 26;
  }

  /* read number of vertices */
  if (read_uint16(file, &gsm.num_vertices)) return 27;
  gsm.vertices = new vertex_t[gsm.num_vertices];

  /* read vertices */
  for (i = 0; i < gsm.num_vertices; i++) {
    vertex_t &vertex = gsm.vertices[i];
    if (read_uint16(file, &vertex.uv_id)) return 28;
    if (vertex.uv_id >= gsm.num_uvs) return 28;
    if (read_uint16(file, &vertex.position_id)) return 29;
    if (vertex.position_id >= gsm.num_positions + gsm.num_blended) return 29;
    if (read_uint16(file, &vertex.stl_id)) return 30;
    if (read_uint16(file, &vertex.hit_detect)) return 31;
  }

  /* read number of triangles */
  if (read_uint16(file, &gsm.num_triangles)) return 32;
  gsm.triangles = new triangle_t[gsm.num_triangles];

  /* read triangles */
  for (i = 0; i < gsm.num_triangles; i++) {
    triangle_t &triangle = gsm.triangles[i];
    if (read_arbitrary(file, &triangle, sizeof(triangle_t))) return 33;
    if (triangle[0] >= gsm.num_vertices) return 33;
    if (triangle[1] >= gsm.num_vertices) return 33;
    if (triangle[2] >= gsm.num_vertices) return 33;
    if (skip(file, 2)) return 34;
  }

  /* read number of groups */
  if (read_uint16(file, &gsm.num_groups)) return 35;
  gsm.groups = new group_t[gsm.num_groups];

  /* read groups */
  for (i = 0; i < gsm.num_groups; i++) {
    group_t &group = gsm.groups[i];
    if (read_uint16(file, &group.material_id)) return 36;
    if (read_uint16(file, &group.first_triangle)) return 37;
    if (read_uint16(file, &group.num_triangles)) return 38;
    if (group.first_triangle >= gsm.num_triangles) return 38;
    if (group.first_triangle + group.num_triangles > gsm.num_triangles) return 38;
    if (skip(file, 2)) return 39;
  }

  /* read number of lists */
  if (read_uint32(file, &gsm.stl_length)) return 40;
  if (read_uint16(file, &gsm.num_stls)) return 41;
  gsm.stls = new stl_t[gsm.num_stls];

  /* read lists */
  for (i = 0; i < gsm.num_stls; i++) {
    stl_t &stl = gsm.stls[i];
    if (read_uint16(file, &stl.num_triangles)) return 42;
    stl.triangles = new unsigned short[stl.num_triangles];
    for (j = 0; j < stl.num_triangles; j++) {
      if (read_uint16(file, &stl.triangles[j])) return 43;
    }
  }
  for (i = 0; i < gsm.num_vertices; i++) {
    vertex_t &vertex = gsm.vertices[i];
    //if (vertex.stl_id >= gsm.num_stls) return 30;
  }

  /* read number of textures */
  if (read_uint16(file, &gsm.num_textures)) return 44;
  if (gsm.num_textures) gsm.textures = new texture_t[gsm.num_textures];

  /* read textures */
  for (i = 0; i < gsm.num_textures; i++) {
    texture_t &texture = gsm.textures[i];
    if (read_uint16(file, &texture.name_length)) return 45;
    texture.name = new char[texture.name_length + 1];
    if (read_arbitrary(file, texture.name, texture.name_length)) return 46;
    texture.name[texture.name_length] = '\0';
    if (skip(file, 2)) return 47;
    set_basename(texture.name, texture.name_length, &texture.basename, &texture.basename_length);
    texture.hidden = 0;
  }

  /* read number of materials */
  if (read_uint16(file, &gsm.num_materials)) return 48;
  gsm.materials = new material_t[gsm.num_materials];

  /* read materials */
  for (i = 0; i < gsm.num_materials; i++) {
    material_t &material = gsm.materials[i];
    if (skip(file, 2)) return 49;
    if (read_uint16(file, &material.name_length)) return 50;
    material.name = new char[material.name_length + 1];
    if (read_arbitrary(file, material.name, material.name_length)) return 51;
    material.name[material.name_length] = '\0';
    if (read_uint16(file, &material.num_maps)) return 52;
    if (material.num_maps) material.maps = new map_t[material.num_maps];
    for (j = 0; j < material.num_maps; j++) {
      map_t &map = material.maps[j];
      if (read_uint16(file, &map.type)) return 53;
      if (read_uint16(file, &map.texture_id)) return 54;
      if (map.texture_id >= gsm.num_textures) return 54;
    }
    if (read_uint16(file, &material.type)) return 55;
    if (read_vec3(file, &material.diffuse)) return 56;
    if (read_vec3(file, &material.ambient)) return 57;
    if (read_vec3(file, &material.specularity)) return 58;
    if (read_float(file, &material.opacity_falloff)) return 59;
    if (read_float(file, &material.transparency)) return 60;
    if (read_char(file, &material.twosided)) return 61;
    if (read_float(file, &material.self_illumination)) return 62;
    if (read_float(file, &material.shininess)) return 63;
    if (read_float(file, &material.shine_strength)) return 64;

    /* make OpenGL values */
    if (material.twosided) material.face = GL_FRONT_AND_BACK;
    else material.face = GL_FRONT;
    material.gl_opacity = 1.0f - material.transparency; /* DEP */
    material.gl_shininess = material.shininess / 100.0f;
    for (k = 0; k < 3; k++) {
      material.gl_diffuse[k] = material.diffuse[k] / 255.0f;
      material.gl_ambient[k] = material.ambient[k] / 255.0f;
      material.gl_specularity[k] = material.specularity[k] / 255.0f;
      material.gl_emission[k] = material.self_illumination / 100.0f;
    }
    material.gl_diffuse[3] = material.gl_ambient[3] = material.gl_specularity[3] = material.gl_emission[3] = 1.0f - material.transparency;
  }

  for (i = 0; i < gsm.num_groups; i++) {
    group_t &group = gsm.groups[i];
    if (group.material_id >= gsm.num_materials) return 36;
  }

  /* grips */
  if (read_grippers(file, &gsm.num_grips, &gsm.grips)) return 65;
  /* and grippers */
  if (read_grippers(file, &gsm.num_grippers, &gsm.grippers)) return 66;

  for (i = 0; i < gsm.num_grips; i++) {
    gripper_t &grip = gsm.grips[i];
    if (grip.bone_id >= gsm.num_bones) return 36;
  }
  for (i = 0; i < gsm.num_grippers; i++) {
    gripper_t &gripper = gsm.grippers[i];
    if (gripper.bone_id >= gsm.num_bones) return 36;
  }

  close_file(file);

  /* classify */
  gsm.type = classify_gsm(gsm);

  return 0;
}

int write_gsm(gsm_t &gsm, char *filename) {
  unsigned short i, j;
  unsigned short zero = 0;
  FILE *file = open_write(filename);
  if (! file) return 1;

  if (write_arbitrary(file, (char *) GSM_MAGIC, 16)) return 2;
  /* write version shit */
  if (write_uint16(file, &gsm.version)) return 2;
  /* skip skeleton shit (short, short) */
  i = 1; if (write_uint16(file, &i)) return 3;
  if (write_uint16(file, &zero)) return 3;
  /* write skeleton crap */
  if (write_uint16(file, &gsm.name_length)) return 3;
  if (write_arbitrary(file, gsm.name, gsm.name_length)) return 4;

  /* write number of bones */
  if (write_uint16(file, &gsm.num_bones)) return 5;

  int ret = write_bones(file, gsm.num_bones, gsm.bones);
  if (ret) return ret;

  /* write number of assignments */
  if (gsm.need_hack) {
    if (write_uint16(file, &gsm.need_hack)) return 12;
  }
  else if (write_uint16(file, &gsm.num_assignments)) return 12;

  /* write assignments */
  for (i = 0; i < gsm.num_assignments; i++) {
    assignment_t &assignment = gsm.assignments[i];
    if (write_uint16(file, &assignment.bone_id)) return 13;
    if (write_uint16(file, &assignment.first_position)) return 14;
    if (write_uint16(file, &assignment.num_positions)) return 15;
    if (write_uint16(file, &zero)) return 16;
  }

  /* blended position header */
  if (gsm.need_hack) {
    i = 0xeeee;
    if (write_uint16(file, &i)) return 17;
    if (write_uint16(file, &gsm.num_positions)) return 17;
    if (write_uint16(file, &gsm.num_blended)) return 17;
    if (write_uint16(file, &zero)) return 21;
  }

  /* write number of positions */
  if (write_uint16(file, &gsm.num_positions)) return 18;

  /* write the positions */
  for (i = 0; i < gsm.num_positions; i++) {
    position_t &position = gsm.positions[i];
    sgVec3 tmp;
    sgCopyVec3(tmp, position.local[3]);
    if (write_vec3(file, &tmp)) return 19;
  }

  /* write number of blended positions */
  if (write_uint16(file, &gsm.num_blended)) return 20;
  if (write_uint32(file, &gsm.blended_length)) return 21;
  /* write blended positions */
  for (i = 0; i < gsm.num_blended; i++) {
    blended_t &blended = gsm.blended[i];
    if (write_uint16(file, &blended.num_positions)) return 22;
    for (j = 0; j < blended.num_positions; j++) {
      if (write_uint16(file, &blended.positions[j])) return 23;
      if (write_float(file, &blended.weights[j])) return 24;
    }
  }

  /* write number of uvs */
  if (write_uint16(file, &gsm.num_uvs)) return 25;

  /* write uvs */
  for (i = 0; i < gsm.num_uvs; i++) {
    uv_t &uv = gsm.uvs[i];
    if (write_arbitrary(file, &uv, sizeof(uv_t))) return 26;
  }

  /* write number of vertices */
  if (write_uint16(file, &gsm.num_vertices)) return 27;

  /* write vertices */
  for (i = 0; i < gsm.num_vertices; i++) {
    vertex_t &vertex = gsm.vertices[i];
    if (write_uint16(file, &vertex.uv_id)) return 28;
    if (write_uint16(file, &vertex.position_id)) return 29;
    if (write_uint16(file, &vertex.stl_id)) return 30;
    if (write_uint16(file, &vertex.hit_detect)) return 31;
  }

  /* write number of triangles */
  if (write_uint16(file, &gsm.num_triangles)) return 32;

  /* write triangles */
  for (i = 0; i < gsm.num_triangles; i++) {
    triangle_t &triangle = gsm.triangles[i];
    if (write_arbitrary(file, &triangle, sizeof(triangle_t))) return 33;
    if (write_uint16(file, &zero)) return 34;
  }

  /* write number of groups */
  if (write_uint16(file, &gsm.num_groups)) return 35;

  /* write groups */
  for (i = 0; i < gsm.num_groups; i++) {
    group_t &group = gsm.groups[i];
    if (write_uint16(file, &group.material_id)) return 36;
    if (write_uint16(file, &group.first_triangle)) return 37;
    if (write_uint16(file, &group.num_triangles)) return 38;
    if (write_uint16(file, &zero)) return 39;
  }

  /* write number of lists */
  if (write_uint32(file, &gsm.stl_length)) return 40;
  if (write_uint16(file, &gsm.num_stls)) return 41;

  /* write lists */
  for (i = 0; i < gsm.num_stls; i++) {
    stl_t &stl = gsm.stls[i];
    if (write_uint16(file, &stl.num_triangles)) return 42;
    for (j = 0; j < stl.num_triangles; j++) {
      if (write_uint16(file, &stl.triangles[j])) return 43;
    }
  }

  /* write number of textures */
  if (write_uint16(file, &gsm.num_textures)) return 44;

  /* write textures */
  for (i = 0; i < gsm.num_textures; i++) {
    texture_t &texture = gsm.textures[i];
    if (write_uint16(file, &texture.name_length)) return 45;
    if (write_arbitrary(file, texture.name, texture.name_length)) return 46;
    if (write_uint16(file, &zero)) return 47;
  }

  /* write number of materials */
  if (write_uint16(file, &gsm.num_materials)) return 48;

  /* write materials */
  for (i = 0; i < gsm.num_materials; i++) {
    material_t &material = gsm.materials[i];
    if (write_uint16(file, &i)) return 49;
    if (write_uint16(file, &material.name_length)) return 50;
    if (write_arbitrary(file, material.name, material.name_length)) return 51;
    if (write_uint16(file, &material.num_maps)) return 52;
    for (j = 0; j < material.num_maps; j++) {
      map_t &map = material.maps[j];
      if (write_uint16(file, &map.type)) return 53;
      if (write_uint16(file, &map.texture_id)) return 54;
    }
    if (write_uint16(file, &material.type)) return 55;
    if (write_vec3(file, &material.diffuse)) return 56;
    if (write_vec3(file, &material.ambient)) return 57;
    if (write_vec3(file, &material.specularity)) return 58;
    if (write_float(file, &material.opacity_falloff)) return 59;
    if (write_float(file, &material.transparency)) return 60;
    if (write_char(file, &material.twosided)) return 61;
    if (write_float(file, &material.self_illumination)) return 62;
    if (write_float(file, &material.shininess)) return 63;
    if (write_float(file, &material.shine_strength)) return 64;
  }

  /* grips */
  if (write_grippers(file, &gsm.num_grips, &gsm.grips)) return 65;
  /* and grippers */
  if (write_grippers(file, &gsm.num_grippers, &gsm.grippers)) return 66;

  close_file(file);

  return 0;
}

char *classify_gsm(gsm_t &gsm) {
  switch (gsm.num_bones) {
    case SWT_NUM_BONES:
      return "officer";
    case KID_NUM_BONES:
      return "child";
    case TAC_NUM_BONES:
      if (gsm.num_grippers) return "male badguy";
      else return "male hostage";
    case BF1_NUM_BONES:
      if (gsm.num_grippers) return "female badguy";
      else return "female hostage or priest";
    case GUN_NUM_BONES:
      if (gsm.num_grips) return "weapon";
    default:
      return "unknown";
  }
}

void cleanup_gsm(gsm_t *gsm) {
  gsm->version = 0;
  if (gsm->name_length && gsm->name) delete[] gsm->name;
  gsm->name_length = 0; gsm->name = 0;
  cleanup_bones(gsm->num_bones, gsm->bones, 0);
  gsm->num_bones = 0; gsm->bones = 0;
  if (gsm->num_assignments && gsm->assignments) delete[] gsm->assignments;
  gsm->num_assignments = 0; gsm->assignments = 0;
  if (gsm->num_positions && gsm->positions) delete[] gsm->positions;
  gsm->num_positions = 0; gsm->positions = 0;
  cleanup_blended(gsm->num_blended, gsm->blended);
  gsm->num_blended = 0; gsm->blended = 0; gsm->blended_length = 0;
  if (gsm->num_uvs && gsm->uvs) delete[] gsm->uvs;
  gsm->num_uvs = 0; gsm->uvs = 0;
  if (gsm->num_vertices && gsm->vertices) delete[] gsm->vertices;
  gsm->num_vertices = 0; gsm->vertices = 0;
  if (gsm->num_triangles && gsm->triangles) delete[] gsm->triangles;
  gsm->num_triangles = 0; gsm->triangles = 0;
  if (gsm->num_groups && gsm->groups) delete[] gsm->groups;
  gsm->num_groups = 0; gsm->groups = 0;
  cleanup_stls(gsm->num_stls, gsm->stls);
  gsm->num_stls = 0; gsm->stls = 0; gsm->stl_length = 0;
  cleanup_textures(gsm->num_textures, gsm->textures);
  gsm->num_textures = 0; gsm->textures = 0;
  cleanup_materials(gsm->num_materials, gsm->materials);
  gsm->num_materials = 0; gsm->materials = 0;
  cleanup_grippers(gsm->num_grips, gsm->grips);
  gsm->num_grips = 0; gsm->grips = 0;
  cleanup_grippers(gsm->num_grippers, gsm->grippers);
  gsm->num_grippers = 0; gsm->grippers = 0;
  cleanup_gl_textures(gsm->gl_textures);
  if (gsm->dirname_length && gsm->dirname) delete[] gsm->dirname;
  gsm->dirname_length = 0; gsm->dirname = 0;
  delete gsm;
}
