#include "ms3d_to_gsm.h"

/* For reordering vertices by joint */
typedef struct {
  unsigned short pre; /* Initial index */
  int sort; /* Thing we are sorting on */
} hash_t;

static hash_t *vertex_mapping;

static int hash_sort(const void *a, const void *b) {
  hash_t *one, *two;

  one = (hash_t *) a;
  two = (hash_t *) b;

  if (one->sort < two->sort) return -1;
  if (one->sort > two->sort) return 1;
  return 0;
}

static bool vector_equiv(sgVec3 &v, sgVec3 &u) {
  for (int i = 0; i < 3; i++) if (fabs(v[i] - u[i]) > 0.0001) return false;
  return true;
}

static int get_parent_joint(ms3d_model_t &ms3d, unsigned short bone_id) {
  if (bone_id >= ms3d.num_joints) return -1;

  ms3d_joint_t &joint = ms3d.joints[bone_id];
  if (! *joint.parent_name) return -1;

  for (unsigned short i = 0; i < ms3d.num_joints; i++) {
    ms3d_joint_t &parent = ms3d.joints[i];
    if (str_equals(parent.name, joint.parent_name)) return (int) i;
  }

  return -1;
}

/* vertex_id is the same as dup_id (dup_id < vertex_id) */
static void mark_dup_vertex(ms3d_model_t &ms3d, unsigned short vertex_id, unsigned short dup_id) {
  unsigned short i, j, k;

  /* Update all triangles to reflect new vertex ID */
  for (i = 0; i < ms3d.num_groups; i++) {
    ms3d_group_t &group = ms3d.groups[i];

    for (j = 0; j < group.num_triangles; j++) {
      ms3d_triangle_t &triangle = ms3d.triangles[group.triangle_id[j]];

      for (k = 0; k < 3; k++) {
        if (triangle.vertex_id[k] == vertex_id) triangle.vertex_id[k] = dup_id;
        else if (triangle.vertex_id[k] > vertex_id) triangle.vertex_id[k]--;
      }
    }
  }

  /* Update vertex array */
  ms3d.num_vertices--;
  ms3d_vertex_t *new_vertices = new ms3d_vertex_t[ms3d.num_vertices];

  for (i = 0; i < ms3d.num_vertices; i++) {
    if (i < vertex_id) j = i;
    else j = i + 1;
    memmove(&new_vertices[i], &ms3d.vertices[j], sizeof(ms3d_vertex_t));
  }

  delete[] ms3d.vertices;
  ms3d.vertices = new_vertices;
}

/* uv_id is the same as dup_id (dup_id < uv_id) */
static void mark_dup_uv(gsm_t &gsm, unsigned short uv_id, unsigned short dup_id) {
  unsigned short i, j;

  /* Update all vertices */
  for (i = 0; i < gsm.num_vertices; i++) {
    vertex_t &vertex = gsm.vertices[i];

    if (vertex.uv_id == uv_id) vertex.uv_id = dup_id;
    else if (vertex.uv_id > uv_id) vertex.uv_id--;
  }

  /* Update UV array */
  gsm.num_uvs--;
  uv_t *new_uvs = new uv_t[gsm.num_uvs];

  for (i = 0; i < gsm.num_uvs; i++) {
    if (i < uv_id) j = i;
    else j = i + 1;
    memmove(&new_uvs[i], &gsm.uvs[j], sizeof(uv_t));
  }

  delete[] gsm.uvs;
  gsm.uvs = new_uvs;
}

/* Cheat and get default bones */
int ms3d_to_gsm_make_default_bone(bone_t &bone, unsigned short bone_id, unsigned short num_bones) {
  if (num_bones == SWT_NUM_BONES) return make_swt_bone(bone, bone_id);
  if (num_bones == KID_NUM_BONES) return make_kid_bone(bone, bone_id);
  if (num_bones == TAC_NUM_BONES) return make_tac_bone(bone, bone_id);
  if (num_bones == BF1_NUM_BONES) return make_bf1_bone(bone, bone_id);
  if (num_bones == 2) return 0;

  fprintf(stderr, "Unknown number of bones (%u): Model probably won't work in SWAT!\n");

  return 0;
}

/* Get default grippers */
int ms3d_to_gsm_make_default_grippers(gsm_t &gsm) {
  unsigned short i;
  int ret;
  int (*make_gripper_function)(gripper_t &, unsigned short);

  gsm.num_grips = 0; gsm.grips = 0;
  gsm.num_grippers = 0; gsm.grippers = 0;

  switch (gsm.num_bones) {
    case SWT_NUM_BONES:
      gsm.num_grippers = SWT_NUM_GRIPPERS;
      make_gripper_function = make_swt_gripper;
      break;
    case KID_NUM_BONES:
      return 0;
    case TAC_NUM_BONES:
      gsm.num_grippers = TAC_NUM_GRIPPERS;
      make_gripper_function = make_tac_gripper;
      break;
    case BF1_NUM_BONES:
      gsm.num_grippers = BF1_NUM_GRIPPERS;
      make_gripper_function = make_bf1_gripper;
      break;
    case 2:
      gsm.num_grips = GUN_NUM_GRIPS;
      gsm.grips = new gripper_t[gsm.num_grips];
      for (i = 0; i < gsm.num_grips; i++) {
        ret = make_gun_grip(gsm.grips[i], i);
        if (ret) {
          fprintf(stderr, "Error code %d making default grip.  This model can't be exported!\n", ret);
          return 1;
        }
        bone_t &bone = gsm.bones[0];
        sgMat4 m;
        m[0][2] = m[1][1] = 1.0f;
        m[2][0] = -1.0f;
        sgMultMat4(gsm.grips[i].local, m, bone.local);
      }
      return 0;
  }

  if (gsm.num_grippers) {
    gsm.grippers = new gripper_t[gsm.num_grippers];
    for (i = 0; i < gsm.num_grippers; i++) {
      ret = make_gripper_function(gsm.grippers[i], i);
      if (ret) {
        fprintf(stderr, "Error code %d making default gripper.  This model can't be exported!\n", ret);
        return 1;
      }
    }
    return 0;
  }

  printf("Don't know how to make grippers for this model.  Sucks to be you.\n");

  return 0;
}

int ms3d_to_gsm(gsm_t &gsm, ms3d_model_t &ms3d) {
  /* Name */
  gsm.name_length = strlen(CONVERTED_NAME);
  gsm.name = new char[gsm.name_length + 1];
  memmove(gsm.name, CONVERTED_NAME, gsm.name_length);
  gsm.name[gsm.name_length] = '\0';

  /* Version */
  gsm.version = 1029;

  /* No blended positions from Milkshape */
  gsm.num_blended = 0; gsm.blended = 0;

  if (ms3d_to_gsm_process_joints(gsm, ms3d)) return 1;
  if (ms3d_to_gsm_process_vertices(gsm, ms3d)) return 2;
  if (ms3d_to_gsm_process_triangles(gsm, ms3d)) return 3;
  if (ms3d_to_gsm_process_materials(gsm, ms3d)) return 4;

  gsm.type = classify_gsm(gsm);

  printf("Converted Milkshape model!\n");
  return 0;
}

int ms3d_to_gsm_process_joints(gsm_t &gsm, ms3d_model_t &ms3d) {
  unsigned short i;

  gsm.num_bones = ms3d.num_joints;
  gsm.bones = new bone_t[gsm.num_bones];

  for (i = 0; i < gsm.num_bones; i++) {
    bone_t &bone = gsm.bones[i];
    ms3d_joint_t &joint = ms3d.joints[i];

    /* Name */
    bone.name_length = strlen(joint.name);
    bone.name = new char[bone.name_length + 1];
    memmove(bone.name, joint.name, bone.name_length);
    bone.name[bone.name_length] = '\0';

    /* Parent */
    int parent = get_parent_joint(ms3d, i);
    if (parent < 0) bone.parent = 0xffff;
    else bone.parent = (unsigned short) parent;

    /* Construct transform matrix */
    int pos_indices[3] = { 0, 1, 2 };
      int rot_indices[3] = { 0, 1, 2 };
#if 0
    //int rot_indices[3] = { 0, 2, 1 };
    //int rot_indices[3] = { 1, 2, 0 };
    //int rot_indices[3] = { 1, 0, 2 };
    //int rot_indices[3] = { 2, 0, 1 };
    //int rot_indices[3] = { 2, 1, 0 };

    joint.rotation[0] *= SG_RADIANS_TO_DEGREES;
    joint.rotation[1] *= SG_RADIANS_TO_DEGREES;
    joint.rotation[2] *= SG_RADIANS_TO_DEGREES;

    //printf("%s: (% 0.4f, % 0.4f, % 0.4f) [% 0.4f, % 0.4f, % 0.4f]\n", joint.name, joint.position[0], joint.position[1], joint.position[2], joint.rotation[0], joint.rotation[1], joint.rotation[2]);
#endif
#if 1
    sgMakeCoordMat4(bone.local, joint.position[pos_indices[0]], joint.position[pos_indices[1]], joint.position[pos_indices[2]], joint.rotation[rot_indices[0]], joint.rotation[rot_indices[1]], joint.rotation[rot_indices[2]]);
#endif
#if 0
    sgCopyVec3(bone.position, joint.position);
    sgEulerToQuat(bone.orientation, -joint.rotation[rot_indices[0]], joint.rotation[rot_indices[1]], joint.rotation[rot_indices[2]]);
    sgQuatToMatrix(bone.local, bone.orientation);
    sgCopyVec3(bone.local[3], bone.position);
#endif

    /* XXX: Cheat */
#if 1
    ms3d_to_gsm_make_default_bone(bone, i, gsm.num_bones);
#endif

    /* Construct world matrix */
    if (bone.parent == 0xffff) sgCopyMat4(bone.world, bone.local);
    else sgMultMat4(bone.world, gsm.bones[bone.parent].world, bone.local);
  }

  /* Create default grippers */
  ms3d_to_gsm_make_default_grippers(gsm);

  return 0;
}

int ms3d_to_gsm_process_vertices(gsm_t &gsm, ms3d_model_t &ms3d) {
  unsigned short i;

  /* Realistically we can't use this model but ... */
  if (! ms3d.num_vertices) return 0;

  /* All vertices must be assigned */
  unsigned short count = 0;
  for (i = 0; i < ms3d.num_vertices; i++) {
    if (ms3d.vertices[i].bone_id < 0) continue;
    count++;
  }
  if (! count) {
    fprintf(stderr, "Milkshape model has no vertices assigned to bones.\n");
    if (ms3d.num_joints) {
      fprintf(stderr, "Assuming all vertices assigned to root.\n");
      for (i = 0; i < ms3d.num_vertices; i++) {
        ms3d.vertices[i].bone_id = 0;
      }
    }
    else {
      fprintf(stderr, "Can't use this model!\n");
      return 1;
    }
  }
  else if (count < ms3d.num_vertices) {
    fprintf(stderr, "Not all vertices are assigned to a bone.  Can't use this model!\n");
    return 1;
  }

  /* First check for dup vertices */
  for (i = 0; i < ms3d.num_vertices; i++) {
    ms3d_vertex_t &v = ms3d.vertices[i];

    /* Look for dups in preceding vertices */
    for (unsigned short j = 0; j < i; j++) {
      ms3d_vertex_t &u = ms3d.vertices[j];
      if (vector_equiv(v.position, u.position)) {
        //printf("%u (%0.1f, %0.1f, %0.1f) (%0.1f, %0.1f, %0.1f)\n", j, v.position[0], v.position[1], v.position[2], u.position[0], u.position[1], u.position[2]);
        mark_dup_vertex(ms3d, i, j); /* i is a dup of j */
        i--;
        break;
      }
    }
  }

  /* We need to reorder the vertices in the model by joint assignment */
  vertex_mapping = new hash_t[ms3d.num_vertices];

  /* This hash has the initial vertex ID and bone ID */
  for (i = 0; i < ms3d.num_vertices; i++) {
    vertex_mapping[i].pre = i; /* pre is original index */
    vertex_mapping[i].sort = ms3d.vertices[i].bone_id; /* Sort on bone ID */
  }

  /* We sort it */
  qsort((void *) vertex_mapping, (size_t) ms3d.num_vertices, sizeof(hash_t), hash_sort);

  /* Make new vertex array */
  ms3d_vertex_t *new_vertices = new ms3d_vertex_t[ms3d.num_vertices];

  for (i = 0; i < ms3d.num_vertices; i++) {
    memmove(&new_vertices[i], &ms3d.vertices[vertex_mapping[i].pre], sizeof(ms3d_vertex_t));
  }

  delete[] ms3d.vertices;
  ms3d.vertices = new_vertices;

  /* And set the new index */
  for (i = 0; i < ms3d.num_vertices; i++) {
    vertex_mapping[i].sort = vertex_mapping[i].pre; /* Sort on original index */
    vertex_mapping[i].pre = i; /* pre is now the sorted index */
  }

  /* Then sort it back so we can query the values */
  qsort((void *) vertex_mapping, (size_t) ms3d.num_vertices, sizeof(hash_t), hash_sort);

  /* Next we replace all vertex identifiers in each triangle */
  for (i = 0; i < ms3d.num_groups; i++) {
    ms3d_group_t &group = ms3d.groups[i];
    for (unsigned short j = 0; j < group.num_triangles; j++) {
      ms3d_triangle_t &triangle = ms3d.triangles[group.triangle_id[j]];
      for (int k = 0; k < 3; k++) {
        if (triangle.vertex_id[k] >= ms3d.num_vertices) {
          fprintf(stderr, "PANIC: g%u/%u t%u/%u v%d: %u > %u\n", i, ms3d.num_groups, j, group.num_triangles, k, triangle.vertex_id[k], ms3d.num_vertices);
          return 1;
        }
        triangle.vertex_id[k] = vertex_mapping[triangle.vertex_id[k]].pre;
      }
    }
  }

  delete[] vertex_mapping;

  /* Create positions ordered by joint assignment */
  gsm.num_positions = ms3d.num_vertices;
  gsm.positions = new position_t[gsm.num_positions];
  gsm.num_assignments = ms3d.num_joints;
  gsm.assignments = new assignment_t[gsm.num_assignments];
  gsm.need_hack = 0;

  char last_bone = -1;
  int assignment_id = -1;
  for (i = 0; i < gsm.num_positions; i++) {
    position_t &position = gsm.positions[i];
    ms3d_vertex_t &vertex = ms3d.vertices[i];
    assignment_t *assignment;

    /* Assignment */
    if (vertex.bone_id != last_bone) {
      last_bone = vertex.bone_id;
      if (++assignment_id >= gsm.num_assignments) {
        fprintf(stderr, "PANIC: Assignment ID out of range!\n");
        return 2;
      }
      else {
        assignment = &gsm.assignments[assignment_id];
        assignment->first_position = i;
        assignment->bone_id = vertex.bone_id;
        assignment->num_positions = 0;
      }
    }
    assignment->num_positions++;

    /* Set bone ID for position */
    position.bone_id = vertex.bone_id;

    /* We have the world position */
    sgMakeIdentMat4(position.world);
    sgCopyVec3(position.world[3], vertex.position);

    /* And must calculate the local */
    bone_t &bone = gsm.bones[position.bone_id];
    sgMat4 bone_world_transpose;
    sgTransposeNegateMat4(bone_world_transpose, bone.world);
    sgMultMat4(position.local, bone_world_transpose, position.world);
  }

  /* Redundant assignments? */
  if (assignment_id == gsm.num_assignments - 1) return 0;

  gsm.num_assignments = assignment_id + 1;
  assignment_t *new_assignments = new assignment_t[gsm.num_assignments];

  for (i = 0; i < gsm.num_assignments; i++) {
    memmove(&new_assignments[i], &gsm.assignments[i], sizeof(assignment_t));
  }

  delete[] gsm.assignments;
  gsm.assignments = new_assignments;

  return 0;
}

int ms3d_to_gsm_process_triangles(gsm_t &gsm, ms3d_model_t &ms3d) {
  unsigned short i, j, k;

  /* Create groups */
  gsm.num_groups = ms3d.num_groups;
  gsm.groups = new group_t[gsm.num_groups];

  unsigned short triangle_id = 0;
  for (i = 0; i < gsm.num_groups; i++) {
    group_t &group = gsm.groups[i];
    ms3d_group_t &mesh = ms3d.groups[i];

    /* Check material */
    if (mesh.material_id < 0) {
      fprintf(stderr, "Group '%s' has no material assigned.  Can't use this model!\n", mesh.name);
      return 1;
    }

    /* Set material and triangle indices */
    group.material_id = mesh.material_id;
    group.first_triangle = triangle_id;
    group.num_triangles = mesh.num_triangles;

    /* Keep track of total number of triangles */
    triangle_id += mesh.num_triangles;
  }

  /* Create triangles */
  gsm.num_triangles = ms3d.num_triangles;
  gsm.triangles = new triangle_t[gsm.num_triangles];

  /* Create one vertex for each triangle vertex */
  gsm.num_vertices = ms3d.num_triangles * 3;
  gsm.vertices = new vertex_t[gsm.num_vertices];

  /* Create one UV for each triangle vertex */
  gsm.num_uvs = ms3d.num_triangles * 3;
  gsm.uvs = new uv_t[gsm.num_uvs];

  /* Go through MS3D triangles and build GSM triangles */
  unsigned short vertex_id = 0;
  unsigned short uv_id = 0;
  triangle_id = 0;
  for (i = 0; i < ms3d.num_groups; i++) {
    ms3d_group_t &mesh = ms3d.groups[i];

    for (j = 0; j < mesh.num_triangles; j++) {
      triangle_t &triangle = gsm.triangles[triangle_id];
      ms3d_triangle_t &t = ms3d.triangles[mesh.triangle_id[j]];

      for (k = 0; k < 3; k++) {
        /* Triangle vertex list */
        if (k == 2) triangle[0] = vertex_id;
        else if (k == 1) triangle[1] = vertex_id;
        else triangle[2] = vertex_id;

        /* Fill in UV */
        uv_t &uv = gsm.uvs[uv_id];
        uv[0] = t.s[k];
        uv[1] = t.t[k];

        /* Set up vertex */
        vertex_t &vertex = gsm.vertices[vertex_id];
        vertex.position_id = t.vertex_id[k];
        vertex.uv_id = uv_id;
        vertex.stl_id = vertex_id; /* Bogus */
        vertex.stl_id = 0;

        /* Keep track of vertices and UVs */
        vertex_id++;
        uv_id++;
      }

      /* Keep track of triangles */
      triangle_id++;
    }
  }

  /* Redundant stuff? */
  int panic = 0;
  if (vertex_id < gsm.num_vertices - 1) panic = 1 | printf("PANIC: Only used %u of %u vertices\n", vertex_id, gsm.num_vertices);
  if (triangle_id < gsm.num_triangles - 1) panic = 1 | printf("PANIC: Only used %u of %u triangles\n", triangle_id, gsm.num_triangles);
  if (panic) return 2;

  /* Find dup UVs */
  for (i = 0; i < gsm.num_uvs; i++) {
    uv_t &u = gsm.uvs[i];

    for (j = 0; j < i; j++) {
      uv_t &v = gsm.uvs[j];

      /* Temporary vectors to do the comparison */
      sgVec3 uu = { u[0], u[1], 0.0f };
      sgVec3 vv = { v[0], v[1], 0.0f };

      if (vector_equiv(uu, vv)) {
        mark_dup_uv(gsm, i, j);
        i--;
        break;
      }
    }
  }

  /* Create bogus shared triangle list */
  gsm.num_stls = 1;
  gsm.stls = new stl_t[gsm.num_stls];
  stl_t &stl = gsm.stls[0];
  stl.num_triangles = gsm.num_triangles;
  stl.triangles = new unsigned short[gsm.num_triangles];
  for (i = 0; i < gsm.num_triangles; i++) stl.triangles[i] = i;
  gsm.stl_length = 2 + gsm.num_triangles;

  /* XXX: Might be dup vertices - unlikely */

  return 0;
}

int ms3d_to_gsm_process_materials(gsm_t &gsm, ms3d_model_t &ms3d) {
  unsigned short i, j;

  /* Create textures */
  gsm.num_textures = ms3d.num_materials;
  gsm.textures = new texture_t[gsm.num_textures];

  /* Create materials */
  gsm.num_materials = ms3d.num_materials;
  gsm.materials = new material_t[gsm.num_materials];

  unsigned short texture_id = 0;
  for (i = 0; i < gsm.num_materials; i++) {
    material_t &material = gsm.materials[i];
    ms3d_material_t &m = ms3d.materials[i];

    /* Set name */
    material.name_length = strlen(m.name);
    material.name = new char[material.name_length + 1];
    memmove(material.name, m.name, material.name_length);
    material.name[material.name_length] = '\0';

    /* Set material type */
    material.type = 3;

    /* Set material properties */
    for (j = 0; j < 3; j++) {
      material.ambient[j] = 255.0f * m.ambient[j] * m.ambient[3];
      material.diffuse[j] = 255.0f * m.diffuse[j] * m.diffuse[3];
      material.specularity[j] = 255.0f * m.specular[j] * m.specular[3];
    }

    material.transparency = 1.0f - m.opacity;
    material.self_illumination = m.emissive[3];
    material.shininess = m.shininess / 128.0f;
    material.shine_strength = 0.05f; /* Default */

    /* make OpenGL values */
    if (material.twosided) material.face = GL_FRONT_AND_BACK;
    else material.face = GL_FRONT;
    material.gl_opacity = 1.0f - material.transparency; /* DEP */
    material.gl_shininess = material.shininess / 100.0f;
    for (j = 0; j < 3; j++) {
      material.gl_diffuse[j] = material.diffuse[j] / 255.0f;
      material.gl_ambient[j] = material.ambient[j] / 255.0f;
      material.gl_specularity[j] = material.specularity[j] / 255.0f;
      material.gl_emission[j] = material.self_illumination / 100.0f;
    }
    material.gl_diffuse[3] = material.gl_ambient[3] = material.gl_specularity[3] = material.gl_emission[3] = 1.0f - material.transparency;

    /* Make texture if needed */
    if (! *m.texture) {
      /* Better fix the texture */
      material.num_maps = 0;
      material.maps = 0;
      continue;
    }

    /* Set filename */
    texture_t &texture = gsm.textures[texture_id];
    texture.name_length = strlen(m.texture);
    texture.name  = new char[texture.name_length + 1];
    memmove(texture.name, m.texture, texture.name_length);
    texture.name[texture.name_length] = '\0';
    texture.hidden = 0;
    set_basename(texture.name, texture.name_length, &texture.basename, &texture.basename_length);

    /* Create map */
    material.num_maps = 1;
    material.maps = new map_t[material.num_maps];

    map_t &map = material.maps[0];
    map.type = 1; /* Diffuse */
    map.texture_id = texture_id;

    /* Keep track of textures */
    texture_id++;
  }

  /* Eliminate redundant textures */
  if (texture_id >= gsm.num_textures) return 0;

  gsm.num_textures = texture_id;
  texture_t *new_textures = new texture_t[gsm.num_textures];

  for (i = 0; i < gsm.num_textures; i++) {
    memmove(&new_textures[i], &gsm.textures[i], sizeof(texture_t));
  }

  delete[] gsm.textures;
  gsm.textures = new_textures;

  /* Leave dups for convenience to modders */

  return 0;
}
